package Boot.MDELite;

import MDLUtilities.Marquee1In_1Out;
import MDLUtilities.Result;
import PrologDB.DB;
import PrologDB.ErrorReport;
import PrologDB.OOSchema;
import PrologDB.SubTableSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;

public class vpl2OOSchema {
    // don -- I think there is a vpl2Schema program in package Boot.allegory;
    // that is virtually identical.  Some maintenance is needed

    private static DB vplDb;
    private static OOSchema oos;
    private static Table vBox, vAsso;
    private static ErrorReport er;
    private static HashSet<String> ll; // for error reporting

    /**
     * Converts X.vpl.pl into X.schema.pl
     *
     * @param args -- single input file "X.vpl.pl"
     */
    public static void main(String... args) {
        // Step 1: standard marquee processing
        Marquee1In_1Out mark = new Marquee1In_1Out(vpl2OOSchema.class,
                ".vpl.pl", ".schema.pl", args);
        String inputFileName = mark.getInputFileName();
        String outputFileName = mark.getOutputFileName();
        String AppName = mark.getAppName();

        // Step 2: open the database and initialize error report
        //         create empty (output) schema
        vplDb = DB.read(inputFileName);
        
        process(vplDb).outputSchema(outputFileName);
    }

    public static Result process(DB vplDb) {
        vBox = vplDb.getTableEH("vBox");
        vAsso = vplDb.getTableEH("vAssociation");

        oos = new OOSchema(vplDb.getName());
        er = new ErrorReport();
        ll = new HashSet<>();
        
        // Step 3: create all tables 
        for (Tuple vb : vBox.tuples()) {
            String id = vb.get("id");
            String name = vb.get("name");
            
            // Step 3.0 prelim error checking
            if (!vb.is("methods","")) {
                er.add("class %s should not have methods",name);
            }
            if (!vb.is("type", "c")) {
                er.add("only classes are permitted in this diagram: see box [%s,%s..}]",
                        id, name);
                continue;
            }

            // Step 3a: create attrstr table schema for it; check later if
            //          there is attrstr table with attrstr duplicate name
            //          also, add identifiers at the end.
            boolean isAbst = vb.getBool("abst");
            TableSchema tsch = new TableSchema(name);
            if (isAbst) tsch.makeAbstract();

            // Step 3b: add all attributes from class
            String attrstr = vb.get("fields");
            if (!attrstr.equals("")) {
                String[] atts = attrstr.split("%");
                for (String s : atts) {
                    tsch.add(s);
                }
            }

            // Step 3c: add attributes from links
            for (Tuple l : vAsso.tuples()) {
                String aid = l.get("id");
                if (l.is("cid1", id) && l.is("arrow2", "BLACK_DIAMOND")) {
                    addLinkColumn(tsch, l, false, false);
                    ll.add(aid);
                    continue;
                }
                if (l.is("cid2", id) && l.is("arrow1", "BLACK_DIAMOND")) {
                    addLinkColumn(tsch, l, true, false);
                    ll.add(aid);
                    continue;
                }
                if (l.is("cid1", id) && l.is("arrow2", "DIAMOND")) {
                    addLinkColumn(tsch, l, false, true);
                    ll.add(aid);
                    continue;
                }
                if (l.is("cid2", id) && l.is("arrow1", "DIAMOND")) {
                    addLinkColumn(tsch, l, true,true);
                    ll.add(aid);
                    continue;
                }
            }

            // Step 3d: now add table schema to database schema and
            //          check for tables with duplicate names
            oos.add(tsch);
        }
        
        // Step 4: now create subtable hash map that contains the 
        //         data for creating subtable declarations
        HashMap<String, LinkedList<String>> subs = new HashMap<>();
        for (Tuple l : vAsso.tuples()) {
            String aid = l.get("id");
            String class1 = getClassName(l, "cid1");
            String class2 = getClassName(l, "cid2");
            if (l.is("arrow1", "TRIANGLE")) {
                if (subs.containsKey(class1)) {
                    LinkedList<String> sublist = subs.get(class1);
                    sublist.add(class2);
                    addAid(aid);
                    continue;
                } else {
                    LinkedList<String> sublist = new LinkedList<>();
                    sublist.add(class2);
                    subs.put(class1, sublist);
                    addAid(aid);
                    continue;
                }
            }
            if (l.is("arrow2", "TRIANGLE")) {
                if (subs.containsKey(class2)) {
                    LinkedList<String> sublist = subs.get(class2);
                    sublist.add(class1);
                    addAid(aid);
                    continue;
                } else {
                    LinkedList<String> sublist = new LinkedList<>();
                    sublist.add(class1);
                    subs.put(class2, sublist);
                    addAid(aid);
                    continue;
                }
            }
        }

        // Step 5: verify that all associations were referenced.
        //         if not, then something is wrong
        for (Tuple l : vAsso.tuples()) {
            String aid = l.get("id");
            if (!ll.contains(aid)) {
                String class1 = getClassName(l, "cid1");
                String class2 = getClassName(l, "cid2");
                er.add("association %s-%s with id=%s was unused -- must have"
                        + " diamond or triangle end", class1, class2, aid);
            }
        }

        // Step 6: now create the subtable declarations
        for (Map.Entry<String, LinkedList<String>> entry : subs.entrySet()) {
            String key = entry.getKey();
            LinkedList<String> lst = entry.getValue();
            SubTableSchema sts = new SubTableSchema(oos.getTableSchemaEH(key));
            for (String c : lst) {
                TableSchema ts = oos.getTableSchema(c);
                if (ts == null) {
                    er.add("unknown table schema %s referenced", c);
                    continue;
                }
                sts.add(ts);
            }
            oos.add(sts);
        }
        
        // Step 7: add identifiers to tables that are not subtables
        oos.addIdentifiersToAllNonSubtables();

        // Step 9: write schema to output file
        er.printEH(System.out);
        return new Result(oos);
    }

    static String getClassName(Tuple l, String whichCid) {
        Tuple t = l.rightSemiJoinFirst(whichCid, vBox, "id");
        if (t == null) {
            er.add("%s of vAssociation tuple with id=%s is non-existent",
                    whichCid, l.get("id"));
            return "WRONG";
        }
        return t.get("name");
    }

    static void addAid(String aid) {
        if (ll.contains(aid)) {
            er.add("association with id=%s has diamond and triangle", aid);
            return;
        }
        ll.add(aid);
    }

    static void addLinkColumn(TableSchema ts, Tuple l, boolean one, boolean openDiamond) {
        String used = one?"role1":"role2";
        String unused = one?"role2":"role1";
        String role = l.get(used);
        
        String cid = one?"cid1":"cid2";
        String className = getClassName(l, cid);
        // if role="" then use the name of the class
        if (role.equals("")) {
            role=className;
        }
        String[] all = role.split(",");
        for (String s : all) {
            if (openDiamond) {
                ts.add(s+":optional "+className);
            } else {
            ts.add(s+":"+className);
            }
        }
        // the unused role should be blank
        if (!l.is(unused,"")) {
            er.add("association with id=%s has an unused role=%s",
                    l.getId(), l.get(unused));
        }
    }
}
