package PrologDB;

import MDLUtilities.Error;
import java.io.PrintStream;
import java.util.LinkedList;
import java.util.List;

/**
 * a SubTableSchema object is an ordered pair (table, [list-of-subtables]),
 * which encodes one level of a table inheritance hierarchy.
 */
public class SubTableSchema {

    /**
     * supr is the schema of belongstoSchema table subs is a list of 
     * schemas that are children of the belongstoSchema table
     */
    private TableSchema supr;
    private final List<TableSchema> subs;
    Schema belongstoSchema = null; // added by Schema.addSubtableSchema()

    // ACCESSORS and INFO
    /**
     * @return subtable schemas of a SubTableSchema declaration
     * MAY BE NULL -- so check! 
     */
    public List<TableSchema> getSubTableSchemas() {
        return subs;
    }

    /**
     * return SubTableSchema of super Table
     *
     * @return SubTableSchema of super Table
     */
    public String getName() {
        return supr.getName();
    }

    /**
     * @return return the TableSchema of super table
     */
    public TableSchema getSuper() {
        return supr;
    }

    /**
     * does this subtable schema include TableSchema s1 as subtable?
     *
     * @param s TableSchema of interest
     * @return true if s1 is a child schema of 'this'
     */
    public boolean contains(TableSchema s) {
        return subs.contains(s);
    }

        /**
     * does this subtable schema include TableSchema with name s as subtable?
     *
     * @param s name of TableSchema of interest
     * @return true if s is a child.
     */
    public boolean contains(String s) {
        return subs.stream().anyMatch(ts->ts.getName().equals(s));
    }
    
    // CONSTRUCTORS
    /**
     * create new SubTable object for belongstoSchema schema supr
     *
     * @param supr is Table object of belongstoSchema schema
     */
    public SubTableSchema(TableSchema supr) {
        this.supr = supr;
        subs = new LinkedList<>();
    }
    
    /**
     * @param supr super table schema
     * @param subs array of sub table schemas
     */
    public SubTableSchema(TableSchema supr, TableSchema... subs) {
        this(supr);
        add(subs);
    }

    /**
     * create subTableSchema declaration for belongstoSchema TableSchema supr 
     * with subtable Schemas subs
     *
     * @param supr TableSchema of belongstoSchema table
     * @param subs list of TableSchemas of children tables
     */
    public SubTableSchema(TableSchema supr, List<TableSchema> subs) {
        this.supr = supr;
        this.subs = subs;
    }

    private SubTableSchema() {
        supr = null;
        subs = new LinkedList<>();
    }
    
    /** create a shallow copy of subtable schema s
     * @param s -- subtable schema to copy
    */
    public SubTableSchema(SubTableSchema s) {
        this.supr = s.supr;
        this.subs = s.subs;
    }

    // CONSTRUCTOR UTILITIES
    /**
     * adds SubTableSchema sub
     *
     * @param sub -- SubTableSchema to add
     */
    public void add(TableSchema sub) {
        subs.add(sub);
    }

    /**
     * adding table schemas in bulk as subtables
     *
     * @param tschemas -- list of table schemas that are subtables of 'this'
     */
    public void add(TableSchema... tschemas) {
        for (TableSchema ts : tschemas) {
            subs.add(ts);
        }
    }
    
    public void remove(TableSchema sub) {
        if (!subs.remove(sub)) {
            String msg = String.format("%s is not a subtable of %s\n",sub.getName(),supr.getName());
            throw new RuntimeException(msg);
        }
    }

    // UTILITIES
    /**
     * return table schema from a list of schemas with the given SubTableSchema
     *
     * @param name -- SubTableSchema of table schema to locate
     * @param list -- in a given list
     * @param schema -- name of the Schema 
     * @return -- tableschema with 'SubTableSchema'
     */
    static TableSchema getTableSchema(String name, List<TableSchema> list, Schema schema) {
        for (TableSchema ts : list) {
            if (ts.getName().equals(name)) {
                return ts;
            }
        }
        return null;
    }

    /** same as TableSchema, except if nothing is found, throw an Error
     * 
     * @param name == 
     * @param list
     * @param schema
     * @return 
     */
    static TableSchema getTableSchemaEH(String name, List<TableSchema> list, Schema schema) {
        TableSchema result = getTableSchema(name, list, schema);
        if (result == null) {
            //String dbname = belongstoSchema==null?"unknown":belongstoSchema.getName();
            throw Error.toss(Error.subTableNoFind, name, schema.getName());
        }
        return result;
    }

    /**
     * return tableschema with given SubTableSchema from 'this'
     *
     * @param name -- SubTableSchema of table schema
     * @return -- tableschema with given 'SubTableSchema'
     */
    public TableSchema getSubTableSchema(String name) {
        for (TableSchema s : getSubTableSchemas()) {
            if (s.getName().equals(name)) {
                return s;
            }
        }
        return null;
    }

    /**
     * copyFrom SubTableSchema by creating a new, pristine version that 
     * literally replicates the data, but is specific to a new database schema.
     *
     * @param newSchema a Schema for cloning a subtable schema decl
     * @return copyFrom of a SubTableSchema w.r.t. newSchema
     */
    public SubTableSchema copy(Schema newSchema) {
        // here's the idea --  we can't just copyFrom a subtableschema -- we must
        // translate its supr and subs into corresponding objects of the new schema.

        // Step 1: create an empty copyFrom and get the list of new subschemas
        SubTableSchema copy = new SubTableSchema();
        List<TableSchema> os = newSchema.getTableSchemas();

        // Step 2: translate supr into supr of new TableSchema
        copy.supr = getTableSchemaEH(supr.getName(), os, belongstoSchema);

        // Step 3: now translate each subschema to the new TableSchema
        for (TableSchema ts : subs) {
            copy.subs.add(getTableSchemaEH(ts.getName(), os, belongstoSchema));
        }

        // Step 4: return true copyFrom
        return copy;
    }

    /**
     * push columns of belongstoSchema into subs
     */
    public void flatten() {
        List<Column> columns = supr.getColumns();
        for (TableSchema t : subs) {
            Column[] reverseColumns = columns.toArray(new Column[columns.size()]);
            t.addFirst(reverseColumns);
        }
    }


    /**
     * two SubTableSchema objects are equal iff they both reference the same
     * superTable, and this.subtables \subsetof ts.subtables, and vice versa
     *
     * @param ts -- SubTableSchema to compare with
     * @return true if both reference identical sets of tables
     */
    public boolean equals(SubTableSchema ts) {
        // Step 1: must reference same belongstoSchema
        if (!ts.getSuper().equals(getSuper())) {
            return false; // dont have the same super
        }

        // Step 2:  ts must reference only subtables in this
        for (TableSchema s1 : ts.getSubTableSchemas()) {
            TableSchema result = getSubTableSchema(s1.getName());
            if (result == null) {
                return false;
            }
        }

        // Step 3: this must reference only subtables in ts
        for (TableSchema s2 : getSubTableSchemas()) {
            TableSchema result = ts.getSubTableSchema(s2.getName());
            if (result == null) {
                return false;
            }
        }

        // if we get this far, the SubTableSchemas are equal
        return true;  // same super SubTableSchema and same subclasses
    }

    /**
     * two SubTableSchema objects are equal iff they both reference the same
     * superTable, and this.subtables \subsetof ts.subtables, and vice versa if
     * unequal, Error is thrown with a reason
     *
     * @param ts -- SubTableSchema to compare with
     */
    public void equalsEH(SubTableSchema ts) {
        // Step 1: must reference same belongstoSchema
        if (!ts.getSuper().equals(getSuper())) {
            String name1 = ts.getSuper().getName();
            String name2 = getSuper().getName();
            if (name1.equals(name2)) {
                throw Error.toss(Error.subTableNoSameSuperSameNames, name1);
            } else {
                throw Error.toss(Error.subTableNoSameSuperDifferentNames, name1, name2);
            }
        }

        // Step 2:  ts must reference only subtables in this
        for (TableSchema s1 : ts.getSubTableSchemas()) {
            TableSchema result = getSubTableSchema(s1.getName());
            if (result == null) {
                String dbname = belongstoSchema==null?"unknown":belongstoSchema.getName();
                throw Error.toss(Error.tableNoExist, s1.getName(), dbname);
            }
        }

        // Step 3: this must reference only subtables in ts
        for (TableSchema s2 : getSubTableSchemas()) {
            TableSchema result = ts.getSubTableSchema(s2.getName());
            if (result == null) {
                String dbname = ts.belongstoSchema==null?"unknown":ts.belongstoSchema.getName();
                throw Error.toss(Error.tableNoExist, s2.getName(), dbname);
            }
        }
        // if we get this far, the SubTableSchemas are equal,return
    }

    // PRINT
    /**
     * produce standard string representation of a SubTableSchema declaration
     *
     * @return standard string representation of a SubTableSchema declaration
     */
    public String toString() {
        String result;
        String comma = "";

        result = String.format("subtable(%s,[", supr.getName());
        for (TableSchema s : subs) {
            result = result + String.format("%s%s", comma, s.getName());
            comma = ",";
        }
        result = result + String.format("]).\n");
        return result;
    }

    /**
     * print standard string declaration of SubTableSchema to PrintStream out
     *
     * @param out is output PrintStream
     */
    public void print(PrintStream out) {
        out.print(this);
    }
    
    /** print to Standard out
     */
    public void print() {
        print(System.out);
    }
}
